/*
 
 Copyright (C) 2010 Timo Lassmann <timolassmann@gmail.com>
 
 This file is part of delve.
 
 Delve is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.
 
 Delve is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
 
 You should have received a copy of the GNU General Public License
 along with Delve.  If not, see <http://www.gnu.org/licenses/>.
 
 */

#include "delve.h"
#include "seed.h"
#include "interface.h"
#include "rand.h"
#include "dbio.h"
#include "fix.h"
#include "read_sam.h"
#include "nuc_code.h"
#include "align_to_sam.h"
#include  <math.h>
#include <string.h>



struct db* fix_sam_file(struct db* db,struct parameters* param)
{
	FILE* file = 0;
	char tmp_text[MAX_LINE];
	char line[MAX_LINE];
	char command[1000];
	//char seq_name[MAX_LINE]; 
	int column = 0; 
	int i,j;
	int tmp = 0;
	unsigned int pos;
	unsigned int add = 0;
	//int read = 0;
	//int c = 0;
	int g = 0;
	int hit = 0;
	struct read_info* ri = 0;
	
	
	ri = malloc(sizeof(struct read_info));
	
	
	
	int storage_num_chr = 10; 
	db = malloc(sizeof(struct db));
	
	db->seq_info = malloc(sizeof(struct db_seq_info));
	db->seq_info->names = malloc(sizeof(char*)* storage_num_chr);
	db->seq_info->sn_len = malloc(sizeof(int)* storage_num_chr);
	db->seq_info->len = malloc(sizeof(unsigned int)* storage_num_chr);
	db->seq_info->nuc_counts = malloc(sizeof(float)*5);
	
	db->bfa = 0;
	db->dbname  = malloc(sizeof(char)*(2));
	db->dbname[0] = 'A';
	db->dbname[1] = 0;
	
	int read_seq_names = 0;
	
	
	unsigned int real_start = 0;
	
	
	int sam = 0;
	int total[11];
	int errors[11];
	
	
	
	int k_distances_in_error[11][5];
	
	int error_types[6][6];
	
	for(i = 0; i < 6;i++){
		for(j = 0; j < 6;j++){
			error_types[i][j] = 0;
		}
	}
	
	int KR[5]; 
	int KA[5];
	for(i =0;i < 5;i++){
		KR[i] = 0;
		KA[i] = 0;
	}
	
	int exclude = 0;
	for(i = 0; i < 11;i++){
		total[i] = 0;
		errors[i] = 0;
		for(j = 0; j < 5;j++){
			k_distances_in_error[i][j] = 0;
		}
	}
	
	
	
	if(!strcmp(".sam", param->infile[1] + (strlen(param->infile[1] ) - 4))){
		sam = 1;
		//}else if (byg_end(".bam", param->infile[i])  == strlen(param->infile[i])){
	}else if (!strcmp(".bam", param->infile[1] + (strlen(param->infile[1] ) - 4))){
		sam = 2;
		//}else if (byg_end(".fa", param->infile[i])  == strlen(param->infile[i])){
	}
		
	
	
	if(!sam){
		fprintf(stderr,"eval only works with sam / bam files.\n");
	}
	if(sam == 2){
		command[0] = 0;
		strcat ( command, "samtools view -H ");     // Copy name into full name
		i = sprintf (tmp_text, "%s ", param->infile[1]);
		strcat ( command, tmp_text);
		//fprintf(stderr,"%s\n",command);
		if (!(file = popen(command, "r"))) {
			fprintf(stderr,"Cannot open bam file '%s' with command:%s\n",param->infile[1],command);
			exit(-1);
		}
	}else{
		if (!(file = fopen(param->infile[1] , "r" ))){
			fprintf(stderr,"Cannot open sam file '%s'\n",param->infile[1]);
			exit(-1);
		}
		
	}
	
	
	
	
	while(fgets(line, MAX_LINE, file)){
		fprintf(stdout,"%s",line);
		if(line[0] == '@'){
			if(line[1] == 'S' && line[2] == 'Q' ){
			
				tmp = byg_end("SN:", line );
				if(tmp){
					db->seq_info->names[read_seq_names] = malloc(sizeof(unsigned char)*(50) );
					add = 0;
					for(j = tmp;j < MAX_LINE;j++){
						if(isspace((int)line[j])){
							db->seq_info->names[read_seq_names][add] = 0;
							break;
						}
						db->seq_info->names[read_seq_names][add] = line[j];
						add++;
					}
				}
				
				//db->seq_info->names[read] = (unsigned char*)strdup(name);
				db->seq_info->sn_len[read_seq_names] = add;
				tmp = byg_end("LN:", line );
				db->seq_info->len[read_seq_names] = atoi(line+tmp);
				read_seq_names++;
				if(read_seq_names == storage_num_chr){
					storage_num_chr = storage_num_chr << 1;
					db->seq_info->names = realloc(db->seq_info->names,sizeof(char*)* storage_num_chr);
					db->seq_info->sn_len = realloc(db->seq_info->sn_len,sizeof(int)* storage_num_chr);
					db->seq_info->len = realloc(db->seq_info->len,sizeof(int)* storage_num_chr);
				}
			}
			
		}else{
			break;
		}
	}
	if(!read_seq_names){
		fprintf(stderr,"SAM header seems to be absent.\n");
		fclose(file);
		return db;
		
	}
	
	db->seq_info->num_seq = read_seq_names;
	//real_start = 0;
	pos = 0;
	for(i = 0; i < db->seq_info->num_seq ;i++){
		real_start = db->seq_info->len[i];
		db->seq_info->len[i] = pos;
		
		pos = real_start + pos;
		
		//fprintf(stderr,"%s	%u\n",db->seq_info->names[i],db->seq_info->len[i]);
	}
	//exit(0);
	//real_start = 0;
	
	if(sam == 2){
		pclose(file);
		command[0] = 0;
		strcat ( command, "samtools view  ");     // Copy name into full name
		i = sprintf (tmp_text, "%s ", param->infile[1]);
		strcat ( command, tmp_text);
		//fprintf(stderr,"%s\n",command);
		if (!(file = popen(command, "r"))) {
			fprintf(stderr,"Cannot open bam file '%s' with command:%s\n",param->infile[1],command);
			exit(-1);
		}
	}else{
		rewind(file);		
	}
	
	
	unsigned char* real_chr_name;
	//int flag;
	
	int until_chr_name = 0;
	int after_pos = 0;
	
	//unsigned int pos;
	unsigned int chr_len = 0;
	

	ri->name = 0;
	ri->qual = 0;
	ri->seq = 0; 
	ri->strand= 0;
	ri->hits = 0;
	ri->priors = 0;
	ri->identity = 0;
	ri->mapq = 0;
	ri->len = 0;
	

	ri->strand = malloc(sizeof (unsigned int)*1);
	ri->hits = malloc(sizeof (unsigned int)*1);
	ri->priors = malloc(sizeof (float)*1);
	ri->identity= malloc(sizeof (float)*1);
	
	
	while(fgets(line, MAX_LINE, file)){
		exclude = 0;
		
		if(ri->name){
			free(ri->name);
			ri->name = 0;
		}
		if(ri->qual){
			free(ri->qual);
			ri->qual = 0;
		}
		if(ri->seq){
			free(ri->seq);
			ri->seq = 0; 
		}
		//if(ri->strand){
		//	free(ri->strand);
		//	ri->strand= 0;
		//}
		//if(ri->hits){
		//	free(ri->hits);
		//	ri->hits = 0;
		//}
		//if(ri->priors){
		//	free(ri->priors);
		//	ri->priors = 0;
		//}
		//if(ri->identity){
		///	free(ri->identity);
		//	ri->identity = 0;
		//}
		ri->mapq = 0;
		
		ri->len = 0;
		
		if(line[0] != '@'){
			column = 1; //<QNAME> 
			tmp = 0;
			hit = 0;
			pos = 0xFFFFFFFFu;
			
			
			for(j = 0;j < MAX_LINE;j++){
				tmp++;
				if(isspace((int)line[j])){
					break;
				}
			}
			
			ri->name = malloc(sizeof(unsigned char)* tmp);
			for(j = 0;j < MAX_LINE;j++){
				
				if(isspace((int)line[j])){
					ri->name[j] = 0;
					break;
				}
				ri->name[j] = line[j];
			}
			
			
			
			for(i = 0; i < MAX_LINE;i++){
				if(line[i] == '\n'){
					break;
				}
				if(isspace((int)line[i])){
					column++;
					switch(column){
						case 2: // <FLAG>
							tmp = atoi(line+i+1);
							ri->strand[hit] = (tmp & 0x10);
							break;
							break;
						case 3: // <RNAME> 
							if(line[i+1] != '*'){
								tmp = 0;
								until_chr_name = i;
								for(j = i+1;j < MAX_LINE;j++){
									if(isspace((int)line[j])){
										tmp_text[tmp] = 0;
										break;
									}
									tmp_text[tmp] = line[j];
									tmp++;
								}
								chr_len = get_chr_offset(db->seq_info, tmp_text);
								if(chr_len == 0xFFFFFFFFu){
									fprintf(stderr,"ERROR: %s not found in indexed sequences\n",tmp_text);
									fprintf(stderr,"%s\n",line);
									exit(-1);
								}
							}
							break;
						case 4: // <POS>
							if(chr_len != 0xFFFFFFFFu){
								//fprintf(stderr,"Adding %u	%u	= %u\n",pos,(unsigned int) atoi(line + i + 1),(unsigned int) atoi(line + i + 1) + pos);
								pos =   (unsigned int) atoi(line + i + 1);
								
							}
							break;
						case 10: // <SEQ>
							
							tmp = 0;
							for(j = i+1;j < MAX_LINE;j++){
								tmp++;
								if(isspace((int)line[j])){
									break;
								}
							}
							
							ri->seq = malloc(sizeof(unsigned char)* tmp);
							g = 0;
							for(j = i+1;j < MAX_LINE;j++){
								
								if(isspace((int)line[j])){
									ri->seq[g] = 0;
									break;
								}
								ri->seq[g] = nuc_code5[(int)line[j]];
								g++;
							}
							
							ri->len = g;
							break;
						case 11: // <QUAL>
							tmp = 0;
							for(j = i+1;j < MAX_LINE;j++){
								tmp++;
								if(isspace((int)line[j])){
									break;
								}
							}
							//if(tmp == 2){
							//	ri[c]->qual = 0;
							//fprintf(stderr,"%d	%s",tmp,line +i+1);
							//}else{
							g= 0;
							ri->qual = malloc(sizeof(unsigned char)* tmp);
							for(j = i+1;j < MAX_LINE;j++){
								
								if(isspace((int)line[j])){
									ri->qual[g] = 0;
									break;
								}
								ri->qual[g] = line[j];
								g++;
							}
							//}
							break;
						default: 
							
							break;
					}
				}
			}
			
			pos = chr_len + pos;
			if(ri->len > 63){
				if(ri->strand[0]){
					ri->seq =  reverse_complement2(ri->seq, ri->len);
				}
				//fprintf(stderr,"Too long...\n%s\n",line);
				unaligned_to_sam(ri);
			}else if(strcmp((char*)get_chr_name( db->seq_info ,pos ) , tmp_text)){
				//unsigned char* get_chr_name(struct db_seq_info* si, unsigned int pos)
				real_chr_name = get_chr_name(db->seq_info,pos);
				fprintf(stderr,"error: should be %s	%d\n%s",real_chr_name ,get_chr_pos(db->seq_info, pos),line);
				
				
				//after_pos = j;
				for(exclude = 0;exclude < until_chr_name;exclude++){
					fprintf(stdout,"%c",line[exclude]);
				}
				fprintf(stdout,"	%s	%d",real_chr_name ,get_chr_pos(db->seq_info, pos));
				
				fprintf(stdout,"%s",line +after_pos );
				
			}else{
				fprintf(stdout,"%s",line);
			}
		}else{
			fprintf(stdout,"%s",line);
		}
		
	}
	//exit(0);
	
	if(sam==2){
		pclose(file);
	}else{
		fclose(file);
	}
	
	if(ri->name){
		free(ri->name);
		ri->name = 0;
	}
	if(ri->qual){
		free(ri->qual);
		ri->qual = 0;
	}
	if(ri->seq){
		free(ri->seq);
		ri->seq = 0; 
	}
	if(ri->strand){
		free(ri->strand);
		ri->strand= 0;
	}
	if(ri->hits){
		free(ri->hits);
		ri->hits = 0;
	}
	if(ri->priors){
		free(ri->priors);
		ri->priors = 0;
	}
	if(ri->identity){
		free(ri->identity);
		ri->identity = 0;
	}
	ri->mapq = 0;
	
	ri->len = 0;
	free(ri);
	//fclose(file);
	
	//fprintf(stderr,"%d\n",	db->seq_info->num_seq );
	
	return db;
}


