/*
 
 Copyright (C) 2010 Timo Lassmann <timolassmann@gmail.com>
 
 This file is part of delve.
 
 Delve is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.
 
 Delve is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
 
 You should have received a copy of the GNU General Public License
 along with Delve.  If not, see <http://www.gnu.org/licenses/>.
 
 */

#include <stdlib.h>
#include <stdio.h>
#include <float.h>
#include <math.h>
#include <pthread.h>
#include "rand.h"
#include "mathsupport.h"
#include "read_sam.h"

#define TAU 0.01f

#define REGIONS 6


#define TAU_LOG prob2scaledprob(TAU)
#define NOT_TAU_LOG prob2scaledprob(1.0f - TAU)

#define MAX(a, b) (a > b ? a : b)
#define MAX3(a,b,c) MAX(MAX(a,b),c)


struct hmm{ 

	
	float** fM;
	float** fC;
	float** fX;
	float** fY;
	float** fR;

	float** bM;
	float** bC;
	float** bX;
	float** bY;
	float** bR;
	
	unsigned short int** traceback;
	
	float* rey;
	
	float*** ec;
	float***em;
	float** ex;
	float** ey;
		
	float*** tem;
	float*** tec;
	float** tex;
	float** tey;
	
	float* mm;
	float* mc;
	float* mgx;
	float* mgy;
	float* mr;
	
	float* cc;
	float* cm;
	float* cgx;
	float* cgy;
	float* cr;
	
	float* gxgx;
	float* gxm;
	float* gxc;
	
	float* gygy;
	float* gym;
	float* gyc;
	float* gyr;
	
	float* rr;
	float* rgy;
	float* rmc;
	
	float* tmm; 
	float* tmc;
	float* tmr;
	float* tmgx;
	float* tmgy;
	
	float* tcc;
	float* tcm;
	float* tcr;
	float* tcgx;
	float* tcgy;

	float* tgxgx;
	float* tgxm;
	float* tgxc;	
	
	float* tgygy;
	float* tgym;
	float* tgyc;
	float* tgyr;
	
	float* trr;
	float* trgy;
	
	float score;
	
	int x;
	int y;
};


struct random_hmm{
	float* ex;
	float* ey;
	float xx;
	float xy;
	float yy;
	float ye;
};

struct thread_data{
	struct read_info** ri;
	struct hmm* hmm;
	struct db* db;
	struct rescue_probabilities* rp;
	struct parameters* param;
	
	int start;
	int end;
};

struct hmm* init_hmm(int x, int y);


void* do_baum_welch_thread(void *threadarg);
void* do_baum_welch_95ID_thread(void *threadarg);
void init_thread_hmms(struct thread_data* thread_data_array,struct hmm* localhmm,int numthreads);
void entangle_hmms(struct thread_data* thread_data_array,struct hmm* localhmm,int numthreads);

void free_hmm(struct hmm* hmm);
void print_hmm_param(struct hmm* hmm);

struct hmm* forward(struct hmm* hmm,unsigned char* a, unsigned char* b, int n,int m);
struct hmm* backward(struct hmm* hmm,unsigned char* a, unsigned char* b, int n,int m);
struct hmm* get_prob(struct hmm* hmm, unsigned char* a, unsigned char* b, int n,int m);
struct hmm* re_estimate(struct hmm* hmm);

unsigned char* global_local_viterbi_log(struct hmm* hmm,float* rey,unsigned char* a, unsigned char* b, int n,int m,int strand);
struct hmm* global_local_forward_log(struct hmm* hmm,float* rey,unsigned char* a, unsigned char* b, int n,int m,int strand);
struct hmm* global_local_backward_log(struct hmm* hmm,float* rey,unsigned char* a, unsigned char* b, int n,int m,int strand);

struct hmm* get_prob_log(struct hmm* hmm, float* rey, unsigned char* a, unsigned char* b, int n,int m,float frac,int strand);

void write_hmm_to_file(struct hmm* hmm,struct parameters * param);
void read_hmm_from_file(struct hmm* hmm,struct parameters * param);





